﻿
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.IdentityModel.User;
using OpcLabs.BaseLib.OperationModel;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace;
using OpcLabs.EasyOpc.UA.AddressSpace.Extensions;
using OpcLabs.EasyOpc.UA.AddressSpace.Standard;
using OpcLabs.EasyOpc.UA.Extensions;
using OpcLabs.EasyOpc.UA.Graphs;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.OpcInteroperabilityTest.UA
{
    partial class UAInteroperabilityWorkshopTest
    {
        #region Static Private Methods

        static private bool AcceptAsArrayReadable(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor,
            [NotNull] IUANodeElement nodeElement)
        {
            if (nodeElement.NodeClass != UANodeClass.Variable)
                return false;

            // ReSharper disable once PossibleNullReferenceException
            if (((byte)(client.ReadValue(endpointDescriptor, nodeElement.NodeId, UAAttributeId.UserAccessLevel)) & 1) == 0)   // IMPROVE: do not hard-code: 1 = Opc.Ua.AccessLevels.CurrentRead
                return false;

            // ReSharper disable once PossibleNullReferenceException
            if ((int)(client.ReadValue(endpointDescriptor, nodeElement.NodeId, UAAttributeId.ValueRank)) != 1)   // IMPROVE: support ValueRank >= 1
                return false;

            var arrayDimensions = (Int64[])client.ReadValue(endpointDescriptor, nodeElement.NodeId, UAAttributeId.ArrayDimensions);
            // ReSharper disable once ConvertIfStatementToReturnStatement
            if (arrayDimensions == null)
                return false;

            return true;
        }

        static private bool AcceptAsArrayWriteable(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor,
            [NotNull] IUANodeElement nodeElement)
        {
            if (nodeElement.NodeClass != UANodeClass.Variable)
                return false;

            // ReSharper disable once PossibleNullReferenceException
            if (((byte)(client.ReadValue(endpointDescriptor, nodeElement.NodeId, UAAttributeId.UserAccessLevel)) & 2) == 0)   // IMPROVE: do not hard-code: 2 = Opc.Ua.AccessLevels.CurrentWrite
                return false;

            // ReSharper disable once PossibleNullReferenceException
            if ((int)(client.ReadValue(endpointDescriptor, nodeElement.NodeId, UAAttributeId.ValueRank)) != 1)   // IMPROVE: support ValueRank >= 1
                return false;

            var arrayDimensions = (Int64[])client.ReadValue(endpointDescriptor, nodeElement.NodeId, UAAttributeId.ArrayDimensions);
            // ReSharper disable once ConvertIfStatementToReturnStatement
            if (arrayDimensions == null)
                return false;

            return true;
        }

        static private bool AcceptAsWriteable(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor,
            [NotNull] IUANodeElement nodeElement)
        {
            if (nodeElement.NodeClass != UANodeClass.Variable)
                return false;

            // ReSharper disable once PossibleNullReferenceException
            // ReSharper disable once ConvertIfStatementToReturnStatement
            if (((byte)(client.ReadValue(endpointDescriptor, nodeElement.NodeId, UAAttributeId.UserAccessLevel)) & 2) == 0)   // IMPROVE: do not hard-code: 2 = Opc.Ua.AccessLevels.CurrentWrite
                return false;

            return true;
        }

        static private void CheckArrayDimensions(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor,
            [NotNull] UANodeDescriptor nodeDescriptor)
        {
            // The ArrayDimensions attribute is actually optional even for arrays.
            Int64[] arrayDimensions = null;
            bool arrayDimensionsRead = false;
            try
            {
                AssertUtilities.WriteLine("Reading array dimensions...");
                arrayDimensions = (Int64[])client.ReadValue(endpointDescriptor, nodeDescriptor, UAAttributeId.ArrayDimensions);
                arrayDimensionsRead = true;
            }
            catch (UAException uaException)
            {
                AssertUtilities.WriteLine($"UA exception (might be OK): {uaException.Message}");
                Exception innerException = uaException.InnerException;
                Assert.IsNotNull(innerException);
                AssertUtilities.WriteLine($"Inner exception (might be OK): {innerException.Message}");
                var statusCodeException = innerException as UAStatusCodeException;
                Assert.IsNotNull(statusCodeException);
                Assert.AreEqual(UACodeBits.BadAttributeIdInvalid, statusCodeException.StatusCode.CodeBits);
                // ok
            }
            AssertUtilities.WriteLine($"Array dimensions read: {arrayDimensionsRead}");

            if (arrayDimensionsRead && (arrayDimensions != null))
            {
                //Assert.IsNotNull(arrayDimensions);
                Assert.AreEqual(1, arrayDimensions.Length, "Count of array dimensions not as expected.");
                Int64 arrayDimension = arrayDimensions[0];
                AssertUtilities.WriteLine($"Array dimension: {arrayDimension}");

                // "A value of 0 for an individual dimension indicates that the dimension has a variable length."
                if ((arrayDimension != 0) && (arrayDimension < 3))
                    AssertUtilities.Inconclusive("Not enough elements");
            }
        }

        #endregion



        #region Test Methods

        [TestMethod]
        public void Group003_AttributeServices_Header____________________()
        {
        }

        /// <summary>
        /// Read() attributes - as supported by the Client - for a standard UA-defined node (e.g. "Server"). Repeat with other nodes.
        /// </summary>
        [TestMethod]
        public void Group003_AttributeServices_Test01_Read_StandardNodes()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
            {
                UAAttributeData attributeData1 = client.Read(endpointDescriptor, UAVariableIds.Server_ServerStatus_State);
                AssertUtilities.WriteLine($"Attribute data 1: {attributeData1}");

                UAAttributeData attributeData2 = client.Read(endpointDescriptor, UAVariableIds.Server_ServerDiagnostics_EnabledFlag);
                AssertUtilities.WriteLine($"Attribute data 2: {attributeData2}");

                UAAttributeData attributeData3 = client.Read(endpointDescriptor, UAVariableIds.Server_ServerStatus);
                AssertUtilities.WriteLine($"Attribute data 3: {attributeData3}");

            }

            TestSection.Assert();
            // IMPROVE: assert
        }

        /// <summary>
        /// Read() all supported attributes for a given vendor node in a single call. Vary the vendor nodes.
        /// </summary>
        [TestMethod]
        public void Group003_AttributeServices_Test02_Read_VendorNodes()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
            {
                int successCount = 0;
                foreach (UANodeDescriptor nodeDescriptor in GetReadNodes(client, endpointDescriptor))
                {
                    Debug.Assert(nodeDescriptor != null);
                    AssertUtilities.WriteLine($"Node descriptor: {nodeDescriptor}");

                    // ReSharper disable once PossibleNullReferenceException
                    UANodeClass nodeClass = (UANodeClass)client.ReadValue(endpointDescriptor, nodeDescriptor, UAAttributeId.NodeClass);
                    AssertUtilities.WriteLine($"Node class: {nodeClass}");
                    IEnumerable<UAAttributeId> attributeIds = nodeClass.GetMandatoryAttributeIds();

                    UAReadArguments[] readArgumentArray = attributeIds
                        .Select(id => new UAReadArguments(endpointDescriptor, nodeDescriptor, id))
                        .ToArray();

                    ValueResult[] valueResultArray = client.ReadMultipleValues(readArgumentArray);
                    for (int i = 0; i < valueResultArray.Length; i++)
                        // ReSharper disable PossibleNullReferenceException
                        AssertUtilities.WriteLine($"Value result ({i}) of attribute {readArgumentArray[i].AttributeId}: {valueResultArray[i]}");
                    // ReSharper restore PossibleNullReferenceException

                    TestSection.Assert();

                    for (int i = 0; i < valueResultArray.Length; i++)
                    {
                        ValueResult valueResult = valueResultArray[i];
                        Assert.IsNotNull(valueResult);

                        if (valueResult.Succeeded)
                        {
                            successCount++;
                            continue;
                        }

                        if (valueResult.Exception is UAStatusCodeException statusCodeException)
                        {
                            int codeBits = statusCodeException.StatusCode.CodeBits;
                            if ((codeBits == UACodeBits.BadOutOfService) ||
                                (codeBits == UACodeBits.BadWaitingForInitialData) ||
                                (codeBits == UACodeBits.Uncertain) ||
                                (codeBits == UACodeBits.UncertainInitialValue))
                                continue;
                        }

                        Assert.Fail($"Result {i} was not acceptable");
                    }

                    if (successCount == 0)
                        AssertUtilities.Inconclusive("No successful results");
                }
            }
        }

        /// <summary>
        /// Read() all supported attributes on a given node while specifying a MaxAge value.
        /// </summary>
        [TestMethod]
        public void Group003_AttributeServices_Test03_Read_MaxAge()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
            {
                UANodeDescriptor nodeDescriptor = GetReadNodes(client, endpointDescriptor).FirstOrDefault();

                Assert.IsNotNull(nodeDescriptor, "Node not found");
                AssertUtilities.WriteLine($"Node descriptor: {nodeDescriptor}");

                // ReSharper disable once PossibleNullReferenceException
                UANodeClass nodeClass = (UANodeClass)client.ReadValue(endpointDescriptor, nodeDescriptor, UAAttributeId.NodeClass);
                AssertUtilities.WriteLine($"Node class: {nodeClass}");
                IEnumerable<UAAttributeId> attributeIds = nodeClass.GetMandatoryAttributeIds();

                UAReadArguments[] readArgumentArray = attributeIds
                    .Select(id => new UAReadArguments(endpointDescriptor, nodeDescriptor, id) { ReadParameters = { MaximumAge = 1000 } })
                    .ToArray();

                ValueResult[] valueResultArray = client.ReadMultipleValues(readArgumentArray);
                for (int i = 0; i < valueResultArray.Length; i++)
                    // ReSharper disable PossibleNullReferenceException
                    AssertUtilities.WriteLine($"Value result of attribute {readArgumentArray[i].AttributeId}: {valueResultArray[i]}");
                // ReSharper restore PossibleNullReferenceException

                TestSection.Assert();

                // ReSharper disable once PossibleNullReferenceException
                Assert.IsFalse(valueResultArray.Any(result => !result.Succeeded), "Some result was not successful");

                // IMPROVE: test the timestamp, against MaxAge?
            }
        }

        /// <summary>
        /// Read() the value attribute while specifying an encoding (not "default"). Vary the encoding.
        /// </summary>
        [TestMethod]
        public void Group003_AttributeServices_Test04_Read_Encoding()
        {
            ClientNotSupported();   // IMPROVE: have this
        }

        /// <summary>
        /// Read() a value that is an array while varying the IndexRange to include specific indexes, the first X, the last X, the middle X. Vary the number of dimensions (e.g. "1:2,1:2" or "1:2,1:2,1:2" etc.) - if available.
        /// </summary>
        // IMPROVE: do more,a s in the test description
        [TestMethod]
        public void Group003_AttributeServices_Test05_Read_Array_IndexRange()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;
            using (var client = new EasyUAClient { Isolated = true })
            {
                UANodeDescriptor nodeDescriptor = GetArrayReadableNode(client, endpointDescriptor);
                AssertUtilities.WriteLine($"Node descriptor: {nodeDescriptor}");

                TestSection.Act();
                CheckArrayDimensions(client, endpointDescriptor, nodeDescriptor);

                UAAttributeData attributeDataWhole = client.Read(endpointDescriptor, nodeDescriptor);
                AssertUtilities.WriteLine($"Attribute data (whole): {attributeDataWhole}");

                const int minimum = 1;
                const int maximum = 2;
                UAIndexRangeList indexRangeList = UAIndexRangeList.OneDimension(minimum, maximum);
                AssertUtilities.WriteLine($"Index range list: {indexRangeList}");
                UAIndexRange indexRange = indexRangeList[0];
                Debug.Assert(indexRange != null);

                UAAttributeData attributeDataRange = client.Read(endpointDescriptor, nodeDescriptor, indexRangeList);
                AssertUtilities.WriteLine($"Attribute data (range): {attributeDataRange}");

                TestSection.Assert();

                Assert.IsTrue(attributeDataRange.HasGoodStatus, "Status should be good");

                object value = attributeDataRange.Value;
                Assert.IsNotNull(value);
                Type valueType = value.GetType();
                Debug.Assert(valueType != null);
                Assert.IsTrue(valueType.IsArray);
                Assert.AreEqual(1, valueType.GetArrayRank());
                var array = (Array)value;
                Assert.AreEqual(1, array.Rank);
                Assert.AreEqual(0, array.GetLowerBound(0));
                Assert.AreEqual(indexRange.Span, array.GetLength(0));
                Assert.AreEqual(indexRange.Span - 1, array.GetUpperBound(0));
            }
        }

        /// <summary>
        /// Write() one or more Value attributes (WriteAccess permitted) for one or more nodes in a single call. Vary nodes.
        /// </summary>
        // IMPROVE: do with more nodes
        [TestMethod]
        public void Group003_AttributeServices_Test06_Write_Value()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            const int value = 0;

            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;
            if (WriteWithUserNameIdentity)
            {
                string userName = UserName;
                Debug.Assert(userName != null, "Missing user name");
                endpointDescriptor.UserIdentity = UserIdentity.CreateUserNameIdentity(userName, Password);
            }

            using (var client = new EasyUAClient { Isolated = true })
            {
                UANodeDescriptor nodeDescriptor = GetWriteableNode(client, endpointDescriptor);
                AssertUtilities.WriteLine($"Node descriptor: {nodeDescriptor}");
                AssertUtilities.WriteLine($"Value: {value}");

                TestSection.Act();
                //object valueRank = client.ReadValue(endpointDescriptor, nodeDescriptor, UAAttributeId.ValueRank);
                //AssertUtilities.WriteLine($"Value rank: {valueRank}");

                client.WriteValue(endpointDescriptor, nodeDescriptor, value);
                // IMPROVE: do more writes in a single call
            }
        }

        /// <summary>
        /// Write() a value that is an array using the IndexRange to vary the number of elements, e.g. all, first X, last X, middle X. Repeat for multi-dimension arrays, if available.
        /// </summary>
        [TestMethod]
        public void Group003_AttributeServices_Test07_Write_Array_IndexRange()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = new UAEndpointDescriptor(EndpointDescriptorCommon);
            if (WriteWithUserNameIdentity)
            {
                string userName = UserName;
                Debug.Assert(userName != null, "Missing user name");
                endpointDescriptor.UserIdentity = UserIdentity.CreateUserNameIdentity(userName, Password);
            }
            using (var client = new EasyUAClient { Isolated = true })
            {
                UANodeDescriptor nodeDescriptor = GetArrayWriteableNode(client, endpointDescriptor);

                TestSection.Act();
                CheckArrayDimensions(client, endpointDescriptor, nodeDescriptor);

                //int[] primingValue = {0, 0, 0};
                //AssertUtilities.WriteLine("Writing priming value...");
                //client.WriteValue(endpointDescriptor, nodeDescriptor, primingValue);

                UAIndexRangeList indexRangeList = UAIndexRangeList.OneDimension(1, 2);
                AssertUtilities.WriteLine($"Index range list: {indexRangeList}");
                int[] value = { 1, 2 };
                AssertUtilities.WriteLine("Writing value...");
                client.WriteValue(endpointDescriptor, nodeDescriptor, value, indexRangeList);
            }
        }

        #endregion



        #region Private Methods

        [CanBeNull]
        private UANodeDescriptor FindArrayReadableNode(
            [NotNull] IEasyUAClient client, 
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            if (ArrayReadableNode1 != null)
                return ArrayReadableNode1;

            IEnumerable<UATreePosition> treePositions = client.TraverseNodeTree(
                endpointDescriptor, UAObjectIds.ObjectsFolder, UABrowseParameters.DataNodes,
                // ReSharper disable once PossibleNullReferenceException
                element => element.NodeId.NamespaceIndex != 0);

            UANodeDescriptor nodeDescriptor = treePositions
                .Take(MaximumSearchNodes)
                // ReSharper disable once AssignNullToNotNullAttribute
                .Where(position => AcceptAsArrayReadable(client, endpointDescriptor, position))
                // ReSharper disable once PossibleNullReferenceException
                .Select(position => position.ToUANodeDescriptor())
                .FirstOrDefault();

            return nodeDescriptor;
        }

        [CanBeNull]
        private UANodeDescriptor FindArrayWriteableNode(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            if (ArrayWriteableNode1 != null)
                return ArrayWriteableNode1;

            IEnumerable<UATreePosition> treePositions = client.TraverseNodeTree(
                endpointDescriptor, UAObjectIds.ObjectsFolder, UABrowseParameters.DataNodes,
                // ReSharper disable once PossibleNullReferenceException
                element => element.NodeId.NamespaceIndex != 0);

            UANodeDescriptor nodeDescriptor = treePositions
                .Take(MaximumSearchNodes)
                // ReSharper disable once AssignNullToNotNullAttribute
                .Where(position => AcceptAsArrayWriteable(client, endpointDescriptor, position))
                // ReSharper disable once PossibleNullReferenceException
                .Select(position => position.ToUANodeDescriptor())
                .FirstOrDefault();

            return nodeDescriptor;
        }

        [CanBeNull]
        private UANodeDescriptor FindWriteableNode(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            if (WriteableNode1 != null)
                return WriteableNode1;

            IEnumerable<UATreePosition> treePositions = client.TraverseNodeTree(
                endpointDescriptor, UAObjectIds.ObjectsFolder, UABrowseParameters.DataNodes,
                // ReSharper disable once PossibleNullReferenceException
                element => element.NodeId.NamespaceIndex != 0);

            UANodeDescriptor nodeDescriptor = treePositions
                .Take(MaximumSearchNodes)
                // ReSharper disable once AssignNullToNotNullAttribute
                .Where(position => AcceptAsWriteable(client, endpointDescriptor, position))
                // ReSharper disable once PossibleNullReferenceException
                .Select(position => position.ToUANodeDescriptor())
                .FirstOrDefault();

            return nodeDescriptor;
        }

        [NotNull]
        private UANodeDescriptor GetArrayWriteableNode(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            UANodeDescriptor nodeDescriptor = FindArrayWriteableNode(client, endpointDescriptor);
            if (nodeDescriptor == null)
                AssertUtilities.Inconclusive("Array writeable node not found");
            AssertUtilities.WriteLine($"Array writeable node: {nodeDescriptor}");
            return nodeDescriptor;
        }

        [NotNull]
        private UANodeDescriptor GetArrayReadableNode(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            UANodeDescriptor nodeDescriptor = FindArrayReadableNode(client, endpointDescriptor);
            if (nodeDescriptor == null)
                AssertUtilities.Inconclusive("Array readable node not found");
            AssertUtilities.WriteLine($"Array readable node: {nodeDescriptor}");
            return nodeDescriptor;
        }

        [NotNull]
        private IEnumerable<UANodeDescriptor> GetReadNodes(
            [NotNull] IEasyUAClient client, 
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            const int count = 5;
            IEnumerable<UATreePosition> treePositions = client.TraverseNodeTree(
                endpointDescriptor, UAObjectIds.ObjectsFolder, UABrowseParameters.DataNodes,
                // ReSharper disable once PossibleNullReferenceException
                element => element.NodeId.NamespaceIndex != 0);

            IEnumerable<UANodeDescriptor> nodesFound = treePositions
                // ReSharper disable once PossibleNullReferenceException
                .Where(position => position.NodeClass == UANodeClass.Variable)
                .Select(position => position.ToUANodeDescriptor());

            // ReSharper disable once AssignNullToNotNullAttribute
            return 
                ReadableNodes.Union(
                    DynamicNodes.Union(
                        ReadableNodes.Select(s => new UANodeDescriptor(s)).Concat(nodesFound).Take(count).ToList()));
        }

        [NotNull]
        private UANodeDescriptor GetWriteableNode(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            UANodeDescriptor nodeDescriptor = FindWriteableNode(client, endpointDescriptor);
            if (nodeDescriptor == null)
                AssertUtilities.Inconclusive("Writeable node not found");
            AssertUtilities.WriteLine($"Writeable node: {nodeDescriptor}");
            return nodeDescriptor;
        }

        #endregion
    }
}
