﻿
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.OperationModel;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace;
using OpcLabs.EasyOpc.UA.AddressSpace.Standard;
using OpcLabs.EasyOpc.UA.AlarmsAndConditions;
using OpcLabs.EasyOpc.UA.Extensions;
using OpcLabs.EasyOpc.UA.Graphs;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.OpcInteroperabilityTest.UA
{
    partial class UAInteroperabilityWorkshopTest
    {
        #region Static Private Methods

        [NotNull]
        static private IEnumerable<UANodeElement> Notifiers(
            [NotNull] IEnumerable<UANodeElement> nodeElements,
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            UANodeElement[] nodeElementArray = nodeElements.ToArray();

            // ReSharper disable AssignNullToNotNullAttribute
            UAReadArguments[] readArgumentsArray = nodeElementArray
                .Select(element => new UAReadArguments(endpointDescriptor, element, UAAttributeId.EventNotifier)).ToArray();
            // ReSharper restore AssignNullToNotNullAttribute

            ValueResult[] valueResultArray = client.ReadMultipleValues(readArgumentsArray);
            for (int i = 0; i < readArgumentsArray.Length; i++)
            {
                ValueResult valueResult = valueResultArray[i];
                Debug.Assert(valueResult != null);

                if (!valueResult.Succeeded)
                    continue;
                object value = valueResult.Value;
                if (!(value is Byte))
                    continue;
                var eventNotifier = (Byte)value;
                bool subscribeToEvents = (eventNotifier & UAEventNotifiers.SubscribeToEvents) != 0;
                if (!subscribeToEvents)
                    continue;

                yield return nodeElementArray[i];
            }
        }

        #endregion



        #region Test Methods

        [TestMethod]
        public void Group008_Eventing_Header____________________()
        {
        }

        /// <summary>
        /// Subscribe to the Server Object's EventNotifier attribute. Specify some basic event fields in the select clause. Verify that events are notified.
        /// </summary>
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void Group008_Eventing_Test01_Subscribe_Server()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
            {
                UANodeDescriptor notifier = UAObjectIds.Server;
                Byte eventNotifier = client.ReadValue<Byte>(endpointDescriptor, UAObjectIds.Server, UAAttributeId.EventNotifier);
                if ((eventNotifier & 0x01) == 0)
                {
                    UANodeElementCollection notifierElements = client.BrowseNotifiers(endpointDescriptor, UAObjectIds.Server);
                    Assert.IsNotNull(notifierElements);

                    int notifierElementsCount = notifierElements.Count;
                    AssertUtilities.WriteLine($"Notifier element count: {notifierElementsCount}");
                    if (notifierElementsCount >= 1)
                    {
                        notifier = notifierElements[0];
                        Assert.IsNotNull(notifier);
                    }
                    else
                    {
                        AssertUtilities.WriteLine("No notifiers found on the Server object, using Server object instead");
                    }
                }
                AssertUtilities.WriteLine($"Notifier: {notifier}");

                UAEventFilter eventFilter = new UAEventFilterBuilder(UABaseEventObject.AllFields);
                var monitoredItemArguments = new EasyUAMonitoredItemArguments(
                    null,
                    endpointDescriptor,
                    notifier,
                    new UAMonitoringParameters(1000, eventFilter))
                {
                    AttributeId = UAAttributeId.EventNotifier,
                    AutoConditionRefresh = true
                };

                int failureCount = 0;
                int successCount = 0;
                int dataCount = 0;
                client.EventNotification += (sender, args) =>
                {
                    Assert.IsNotNull(args);
                    AssertUtilities.WriteLine(args);
                    if (args.Succeeded)
                    {
                        successCount++;
                        UAEventData eventData = args.EventData;
                        if (eventData != null)
                        {
                            dataCount++;
                            foreach (KeyValuePair<UAAttributeField, ValueResult> pair in eventData.FieldResults)
                                AssertUtilities.WriteLine($"{pair.Key} -> {pair.Value}");
                        }
                    }
                    else
                        failureCount++;
                };
                int handle = client.SubscribeMonitoredItem(monitoredItemArguments);

                Thread.Sleep(20 * 1000);

                client.UnsubscribeMonitoredItem(handle);

                TestSection.Assert();

                AssertUtilities.WriteLine($"Data count: {dataCount}");
                AssertUtilities.WriteLine($"Success count: {successCount}");
                AssertUtilities.WriteLine($"Failure count: {failureCount}");

                Assert.IsTrue(dataCount >= 1, "Not enough data");
                Assert.IsTrue(successCount >= 2, "Not enough succeeded");
                Assert.AreEqual(0, failureCount, "Some failures");
            }
        }

        /// <summary>
        /// Subscribe with a filter for EventType, Priority etc. Verify that events match the filter criteria.
        /// </summary>
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void Group008_Eventing_Test02_Subscribe_Filter()
        {
            ClientNotSupported();   // IMPROVE: have this
        }

        /// <summary>
        /// [obsolete] Publish Service delivers event notifications matching the filter.
        /// </summary>
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void Group008_Eventing_Test03_Publish_Notifications()
        {
            ClientNotSupported();   // IMPROVE: have this
        }

        [TestMethod]
        public void Group008_Eventing_Test04_EventNodes_Browse()
        {
            TestSection.Arrange();
            const int count = 100;
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
            {
                AssertUtilities.Phase("Via references");

                var eventSources = new UABrowseParameters(UANodeClass.All);
                eventSources.ReferenceTypeIds.Add(UAReferenceTypeIds.HasEventSource);

                IEnumerable<UATreePosition> treePositions1 = client.TraverseNodeTree(
                    endpointDescriptor,
                    UAObjectIds.Server,
                    eventSources);
                int i = 0;
                foreach (UATreePosition treePosition in treePositions1.Take(count))
                {
                    // ReSharper disable once PossibleNullReferenceException
                    AssertUtilities.WriteLine($"[{i}] ({treePosition.Level}) {treePosition.DisplayName}: {treePosition}");
                    i++;
                }
                Assert.IsTrue(i >= 1, "Not enough elements");

                AssertUtilities.Phase("Via notifiers");

                IEnumerable<UATreePosition> treePositions2 = client.TraverseNodeTree(
                    endpointDescriptor,
                    UAObjectIds.ObjectsFolder,
                    UABrowseParameters.Objects); // IMPROVE: what about Views?
                i = 0;
                foreach (UANodeElement nodeElement in Notifiers(treePositions2.Take(count), client, endpointDescriptor))
                {
                    // ReSharper disable once PossibleNullReferenceException
                    AssertUtilities.WriteLine($"{nodeElement}");
                    i++;
                }
                Assert.IsTrue(i >= 1, "Not enough elements");
            }
        }

        #endregion
    }
}
